from __future__ import annotations

access_log_format = "bob"
bind = ["127.0.0.1:5555"]
certfile = "tests/assets/cert.pem"
ciphers = "ECDHE+AESGCM"
h11_max_incomplete_size = 4
keyfile = "tests/assets/key.pem"
